# Copyright (C) 2023 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Callable
from tidalapi.album import Album as TidalAlbum


class AlbumSortCriteria:

    def __init__(self, extractor: Callable[[TidalAlbum], any], comparator: Callable[[any, any], int]):
        self._extractor: Callable[[TidalAlbum], any] = extractor
        self._comparator: Callable[[any, any], int] = comparator

    def compare(self, left: TidalAlbum, right: TidalAlbum) -> int:
        return self._comparator(self._extractor(left), self._extractor(right))
